-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

M.enabled = true

M.dynamicAxises = {
    axisx0 = 3, -- steer
    axisy0 = 1, -- throttle
    axisy1 = 2  -- brake
}

M.rateMult = 1

local throttleStartRate = 1
local throttleEndRate = 0.1
local throttleInRate = 310
local throttleOutRate = 310
local throttleRateApply = 1
local throttleValue = 0

local brakeStartRate = 1
local brakeEndRate = 0.05
local brakeInRate = 310
local brakeOutRate = 580
local brakeRateApply = 0.8
local brakeValue = 0

local steerInRate = 2.5
local steerOutRate = 1.25
local steerAutoCenterRate = 1
local steerValue = 0

--local steerSmoothing = newExponentialSmoothing(5)
local slidingSmoothing = newExponentialSmoothing(5)
local steeringLockLimitSmoothing = newExponentialSmoothing(5)

local tractionSmoothingFront = newExponentialSmoothing(5)
local tractionSmoothingRear = newExponentialSmoothing(5)

local simpleTractionEnabled = false

local function getSlip()
    local wheelRLslip
    local wheelRRslip
	
	local sliding

    --check how much it's sliding
    for i,wd in pairs(drivetrain.wheels) do
        if wd.name == "RL" then
            wheelRLslip = drivetrain.wheelInfo[i].localLastSlip
        elseif wd.name == "RR" then
            wheelRRslip = drivetrain.wheelInfo[i].localLastSlip
        end
    end

	if wheelRLslip and wheelRRslip then
		sliding = math.abs((wheelRLslip + wheelRRslip)/2)
	else
		sliding = 0
	end
    --smooth the inherently unstable sliding value
    sliding = slidingSmoothing:get(sliding)
	
	return sliding
end

local function throttleSmoothing( sample, dt)
    local revState
    local percentIncrease
    local rate
    local state

    local wheelFL
    local wheelFR
    local wheelRL
    local wheelRR
    local wheelSpeedFront = 1
    local wheelSpeedRear = 1
    local tractionEnabled = false

    -- simple traction control
    local tractionRate = 1

    if table.getn(drivetrain.wheels) > 2 then
        for i,wd in pairs(drivetrain.wheels) do
            if wd.name == "FL" then
                wheelFL = obj:getWheel(i)
            elseif wd.name == "FR" then
                wheelFR = obj:getWheel(i)
            elseif wd.name == "RL" then
                wheelRL = obj:getWheel(i)
            elseif wd.name == "RR" then
                wheelRR = obj:getWheel(i)
            end
        end

		wheelSpeedFront = tractionSmoothingFront:get(math.max(math.abs(wheelFL.angularVelocity), math.abs(wheelFR.angularVelocity)))
		wheelSpeedRear = tractionSmoothingRear:get(math.max(math.abs(wheelRL.angularVelocity), math.abs(wheelRR.angularVelocity)))
    end
	
	local sliding = getSlip()

    if wheelSpeedFront / wheelSpeedRear < 0.6 then
        tractionEnabled = true
        tractionRate = 0.25
        sample = 0
    end
	
	if sliding > 3 then
		tractionEnabled = true
		tractionRate = 1
		sample = 0
	end


    if sample == 1 then
        state = throttleValue / throttleRateApply
        revState = math.max(0, 1 - state)
        rate = throttleInRate * dt
        percentIncrease = ( ( throttleStartRate - throttleValue ) * revState ) + throttleEndRate

        throttleValue = throttleValue + ( percentIncrease * rate / 100 )
        return throttleValue
    else
        revState = throttleValue
        rate = throttleOutRate * dt * tractionRate
        percentIncrease = ( ( throttleStartRate - throttleEndRate ) * revState ) + throttleEndRate
        throttleValue = math.max(0, throttleValue - ( percentIncrease * rate / 100 ))

        if tractionEnabled then
            return throttleValue
        else
            return 0
        end
    end
end

local function brakeSmoothing( sample, dt)
    local revState
    local percentIncrease
    local rate
    local state

    if sample == 1 then
        state = brakeValue / brakeRateApply
        revState = math.max(0, 1 - state)
        rate = brakeInRate * dt
        percentIncrease = ( ( brakeStartRate - brakeValue ) * revState ) + brakeEndRate

        brakeValue = brakeValue + ( percentIncrease * rate / 100 )
        return brakeValue
    else
        revState = brakeValue
        rate = brakeOutRate * dt
        percentIncrease = ( ( brakeStartRate - brakeEndRate ) * revState ) + brakeEndRate
        brakeValue = math.max(0, brakeValue - ( percentIncrease * rate / 100 ))

        return 0
    end
end

local function steerSmooth(sample, dt, rateModifier, steeringLockLimit)
    local rate
    local dir = (sample - steerValue) * sign(steerValue)

    -- sample: left = -1; right = 1
    if sample ~= 0 then
        if sample > 0 then
            if dir >= 0 then
                rate = steerOutRate * M.rateMult * rateModifier
            else
                rate = steerInRate * M.rateMult * rateModifier
            end
        else
            if dir <= 0 then
                rate = steerInRate * M.rateMult * rateModifier
            else
                rate = steerOutRate * M.rateMult * rateModifier
            end
        end
    else
        rate = steerAutoCenterRate * M.rateMult * rateModifier
    end

    if sample < steerValue then
        steerValue = math.max((steerValue - dt * rate), sample * steeringLockLimit)
    else
        steerValue = math.min((steerValue + dt * rate), sample * steeringLockLimit)
    end

    steerValue = math.max(math.min(steerValue, 1 * steeringLockLimit), -1 * steeringLockLimit)
    return steerValue
end

local function dynamicSteering( sample, dt )
    local tmp
    local rateModifier = 1
    --get speed
    local airSpeed = math.abs(obj:getVelocity():length())

    local steeringLockLimit = math.abs(225/((airSpeed + 1) * (airSpeed + 1)))

    if steeringLockLimit < 0.20 then steeringLockLimit = 0.20 end
    if steeringLockLimit > 1 then steeringLockLimit = 1 end

	sliding = getSlip()
    --counter steer special case - increase lock with rear wheel slip
    if sliding > 3 then
        sliding = math.min( 15, sliding - 3) / 15
        rateModifier = sliding + 0.4
        steeringLockLimit = steeringLockLimit + sliding
    end

    steeringLockLimit = steeringLockLimitSmoothing:get(steeringLockLimit)

    if steeringLockLimit < 0.20 then steeringLockLimit = 0.20 end
    if steeringLockLimit > 1 then steeringLockLimit = 1 end

    tmp = steerSmooth( sample, dt, rateModifier * steeringLockLimit, steeringLockLimit )

    return tmp
end

local function getInput( axis, sample, dt )
    if axis == "axisy0" then
        return throttleSmoothing( sample, dt )
    elseif axis == "axisy1" then
        return brakeSmoothing( sample, dt )
    elseif axis == "axisx0" then
        return dynamicSteering( sample, dt )
    else
        print("No handler in dynamic keyboard mod for " .. axis)
    end
end

local function isDynamicAxis( axis )
    local isdynamic = false

    for k, v in pairs(M.dynamicAxises) do
        if k == axis then
            isdynamic = true
        end
    end

    return isdynamic
end

local function isEnabled()
    return M.enabled
end

local function setState( state )
    M.enabled = state
end

M.getInput = getInput
M.isEnabled = isEnabled
M.setState = setState
M.isDynamicAxis = isDynamicAxis

return M